/* 
  Copyright (c) 2013 - 2014, Oracle and/or its affiliates. All rights reserved.
  
  This program is free software; you can redistribute it and/or modify it under 
  the terms of the GNU General Public License as published by the Free Software 
  Foundation; version 2 of the License.
  
  This program is distributed in the hope that it will be useful, but WITHOUT ANY 
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS 
  FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License along with this 
  program; if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth 
  Floor, Boston, MA 02110-1301  USA
*/
#pragma once

#ifndef JCALLBACK
#define JCALLBACK

#include "jstack.h"
#include "jkey.h"
#include "jutil.h"

#include <string.h>
#include <stdio.h>
#include <stdlib.h>

/* 
  Pure virtual class for callbacks.
  Must be overloaded or each of functions:
  JCallback_valid, JCallback_extract_key and so on.
*/

class JCallback
{
protected:
  //stack_jelem jstack[MAX_JSON_DEPTH];         /* stack of key=>value pairs */
  stack_jelem* jstack;
  /* 
  For the JSON "foo = { bar = { baz = "blah" } } }" when parser has parsed
  "blah" stack will look like:
  [
    { key_name: "foo =..."; key_len: 3; value: "{ bar..." },
    { key_name: "bar =..."; key_len: 3; value: "{ baz..." },
    { key_name: "baz =..."; key_len: 3; value: "blah\"..." }
  ]
  */
  int num_elem;                               /* number of elements in the stack */
  
public:
  /* 
    Constructor, initializes members
  */
  
  JCallback();
  virtual ~JCallback();
  
  /* 
    Virtual callback function.
    Must be overloaded in a child class.
    
    @param[in]  key_start       start position of the key
    @param[in]  key_len         length of the key
    @param[in]  val_start       start position of the value
    @param[in]  val_len         length of the value
    @param[in]  pos             key position in the tree
    @param[in]  is_array_elem   is this element of an array?
    @param[in]  type_flag       enum_jtype of the element
    
    @return Operation status
      @retval true  OK
      @retval false Failure
  */
  
  
  virtual bool callback_func(int key_start, int key_len, int val_start, int val_len, int pos, bool is_array_elem = false, int type_flag= jdefault) = 0;
  
protected:
  /* 
    Puts element in the top of the stack
    
    @param[in]  key_start       start position of the key
    @param[in]  key_len         length of the key
    @param[in]  val_start       start position of the value
    @param[in]  val_len         length of the value
    @param[in]  type_flag       type of the value
    
    @return Operation status
      @retval true  OK
      @retval false Failure
  */
  
  
  bool put_elem(int key_start, int key_len, int val_start, int val_len, int type_flag= jdefault);
  
  
  /* 
    Checks if found values are in proper order
    
    @param[in]  start_ind   from which index to start
    
    @return if values are in order
      @retval true  in order
      @retval false mixup
  */
  bool check_order(int start_ind= 0);
};

class JCallback_valid: public JCallback
{
  bool callback_func(int key_start, int key_len, int val_start, int val_len, int pos, bool is_array_elem = false, int type_flag= jdefault);
};


/* 
  Calback for parser tests only.
  Not supposed to be used in production
*/
class JCallback_test_parser: public JCallback
{
  bool callback_func(int key_start, int key_len, int val_start, int val_len, int pos, bool is_array_elem = false, int type_flag= jdefault);
  
public:
  char* get_parse_tree(char* doc);
};


class JCallback_contains_key: public JCallback
{
  bool callback_func(int key_start, int key_len, int val_start, int val_len, int pos, bool is_array_elem = false, int type_flag= jdefault);
  
protected:
  char* doc;
  JKey** keys;
  int klen;
  int ind;
  int last_proceed_ind;
  int* arr_inds;
  int arr_ind;
  bool key_exists;
  int failed_pos;
  bool proceed_further;
  bool proceeding_array;
  int skip_stack_elem;
  int prev_pos;
  stack_jelem** jstack_arrays;    /* array to store elements for negative indexes */
  
  bool do_callback(int key_start, int key_len, int val_start, int val_len, int pos, bool is_array_elem = false, int type_flag= jdefault);
  
  /* 
    Puts element in the top of the array stack
    
    @param[in]  cur_pos         current position in the keys array
    @param[in]  arr_ind         current array index
    @param[in]  key_start       start position of the key
    @param[in]  key_len         length of the key
    @param[in]  val_start       start position of the value
    @param[in]  val_len         length of the value
    @param[in]  type_flag       type of the value
    
    @return Operation status
      @retval true  OK
      @retval false Failure
  */
  
  
  bool put_array_elem(int cur_pos, int arr_ind, int key_start, int key_len, int val_start, int val_len, int type_flag= jdefault);
  
public:
  JCallback_contains_key(char* doc, JKey** keys, int klen);
  virtual ~JCallback_contains_key();
  bool contains();
};


class JCallback_extract: public JCallback_contains_key
{
  char* value;
  
  bool callback_func(int key_start, int key_len, int val_start, int val_len, int pos, bool is_array_elem = false, int type_flag= jdefault);
  
public:
  JCallback_extract(char* doc, JKey** keys, int klen);
  char* extract();
};


/*
  Parent class for json_replace, json_append and json_set
*/
class JCallback_remove: public JCallback_contains_key
{
  bool callback_func(int key_start, int key_len, int val_start, int val_len, int pos, bool is_array_elem = false, int type_flag= jdefault);
  
protected:
  int key_start;
  int key_len;
  int val_start;
  int val_len;
  
  void prepare_doc(bool change_key= false, int val_offset= 0);
  
public:
  JCallback_remove(char* doc, JKey** keys, int klen);
  char* get_doc();
};



class JCallback_replace: public JCallback_remove
{
  char* new_value;
  
public:
  JCallback_replace(char* doc, JKey** keys, int klen, char* new_value);
  char* get_doc();
};



/*
  Parent for json_set
*/
class JCallback_append: public JCallback_replace
{
protected:
  char* new_value;
  bool last_found;
  bool maximum_passed;
  bool in_array;
  int decrement;
  bool json_set;
  bool elem_found;
  //int prev_pos;
  bool last_is_empty;
  
  bool callback_func(int key_start, int key_len, int val_start, int val_len, int pos, bool is_array_elem = false, int type_flag= jdefault);
  
protected:
  int prepare_value();
  
public:
  JCallback_append(char* doc, JKey** keys, int klen, char* new_value);
  char* get_doc();
};


class JCallback_set: public JCallback_append
{
public:
  JCallback_set(char* doc, JKey** keys, int klen, char* new_value);
  char* get_doc();
};


class JCallback_search: public JCallback
{
  char* doc;
  char* value;
  JPath** path;
  int path_len;
  int cur_pos;
  bool exists;
  int arr_ind;
  int search_pos;   /* Which occurrence to search? Default: first. */
  
  bool callback_func(int key_start, int key_len, int val_start, int val_len, int pos, bool is_array_elem = false, int type_flag= jdefault);
  
public:
  JCallback_search(char* doc, char* value, int search_pos= 1);
  ~JCallback_search();
  char* get_path(bool path_format= false);
};


class JCallback_depth: public JCallback
{
  int depth;
  
  bool callback_func(int key_start, int key_len, int val_start, int val_len, int pos, bool is_array_elem = false, int type_flag= jdefault);
  
public:
  JCallback_depth();
  int get_depth();
};


class JCallback_count: public JCallback
{
  int count;
  
  bool callback_func(int key_start, int key_len, int val_start, int val_len, int pos, bool is_array_elem = false, int type_flag= jdefault);
  
public:
  JCallback_count();
  int get_count();
};


/* Class, serving JSON_DEEP_MERGE function */
class JCallback_merge: public JCallback
{
  Jtree_elem_childs* cur;       /* Current element in the tree */
  Jtree_elem* root;             /* Tree root */
  int save_pos;                 /* Position, processed at the previous step */
  int hier;                     /* Hierarchy level */
  int pos_in_cur;               /* Position in cur array */
  char* doc;                    /* Pointer to doc */
  Jtree_elem** garbage;         /* Garbage collection for unused curs */
  int garbage_pos;              /* Current element in garbage collection */
  
  
  bool callback_func(int key_start, int key_len, int val_start, int val_len, int pos, bool is_array_elem = false, int type_flag= jdefault);
  void clean_root(Jtree_elem* root);
  
public:
  JCallback_merge(char* doc);
  ~JCallback_merge();
  Jtree_elem* get_merge_tree();   /* Returns tree root */
  stack_jelem* get_stack();       /* Returns stack */
};

#endif /* JCALLBACK */
/*
* :tabSize=2:indentSize=2:noTabs=true:
* :folding=custom:collapseFolds=1:
*/
