/* 
  Copyright (c) 2013 - 2014, Oracle and/or its affiliates. All rights reserved.
  
  This program is free software; you can redistribute it and/or modify it under 
  the terms of the GNU General Public License as published by the Free Software 
  Foundation; version 2 of the License.
  
  This program is distributed in the hope that it will be useful, but WITHOUT ANY 
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS 
  FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License along with this 
  program; if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth 
  Floor, Boston, MA 02110-1301  USA
*/
#include "JParser.h"

JParser::JParser(char* document, JCallback* callback_function)
{
  this->doc=              document;
  this->callback=         callback_function;
  this->cur_type=         jdefault;
  this->elems=            0;
  this->stop_parsing=     false;
  this->parent_types=     new enum_jtype[MAX_JSON_DEPTH + 1];
  memset(this->parent_types, jdefault, sizeof(enum_jtype) * (MAX_JSON_DEPTH + 1)); 
  this->key_start=        new int[MAX_JSON_DEPTH + 1];
  this->key_start[0]=     0;
  this->key_end=          new int[MAX_JSON_DEPTH + 1];
  this->key_end[0]=       0;
  this->val_start=        new int[MAX_JSON_DEPTH + 1];
  this->val_start[0]=     0;
  this->val_end=          new int[MAX_JSON_DEPTH + 1];
  this->val_end[0]=       0;
  this->array_comma=      false;
  this->is_key=           false;
  this->switch_to_pair=   false;
  this->parse_error=      false;
  this->waiting_closure=  false;
};


JParser::~JParser()
{
  delete[] this->parent_types;
  delete[] this->key_start;
  delete[] this->key_end;
  delete[] this->val_start;
  delete[] this->val_end;
  
  this->elem_keys.clear();
};


/* 
  JSON Parser.
  
  Parsers element and when finds a valid JSON pair, calls callback function.
  Depending from its return value either continue parsing until end of the document
  or stops parsing. This allows to save time, say, if we found required element.
  
  Parsing starts from parsing single char:
  
  while (char)
  {
    parsechar(char)
  }
  
  Then, depending from the current object we are in, specific parser function
  is called.
  
  bool parsechar(char)
  {
    switch (state)
    {
      case 'jarray':
        parse_array(char);
        break;
      case 'jstring':
        parse_string(char);
        break;
      ...
      default:
        parse_default(char);
        break;
      }
    }
    
    Then parse_array, parse_string, etc. are states,
    so we don't have thousands of if-else
*/
bool JParser::parse()
{
  int size= strlen(doc);
  
  for (int i= 0; i < size; i++)
  {
    if (!parse_char(doc[i], i) || parse_error)
    {
      jdump(JSON_ERROR, JSON_PARSE_ERROR, "document proceeded up to %d'th character %c", i, doc[i]);
      return false;
    }
    if (stop_parsing)
      return true;
  }
  
  if (jdefault != cur_type
    || 0 != elems
    )
    return false;
  else
    return true;
};


bool JParser::parse_char(char c, int pos)
{
  switch (cur_type)
  {
    if (waiting_closure)
      return false;
  case jobject:
    return parse_object(c, pos);
  case jarray:
    return parse_array(c, pos);
  case jpair:
    return parse_pair(c, pos);
  default:
    if (waiting_closure)
      return false;
    switch (cur_type)
    {
    case jstring:
      return parse_string(c, pos);
    case jtrue:
      return parse_true(c, pos);
    case jfalse:
      return parse_false(c, pos);
    case jnull:
      return parse_null(c, pos);
    case jnum:
      return parse_num(c, pos);
    case jutf:
      return parse_utf_escape(c, pos);
    default:
      return parse_default(c, pos);
    }
  }
}


bool JParser::parse_default(char c, int pos)
{
  switch (c)
  {
  case '{':
    if (is_key || jarray == cur_type || 0 == elems)
      is_key=               false;
    else
    {
      return false;
    }
    parent_types[elems]=  cur_type;
    val_start[elems]=     pos;
    cur_type=             jobject;
    array_comma=          false;  //Dangerous: can mix with prev. array ! Fix it later.
    elems++;
    break;
  case '[':
    if (is_key || jarray == cur_type || 0 == elems)
      is_key=               false;
    else
      return false;
    parent_types[elems]=  cur_type;
    val_start[elems]=     pos;
    cur_type=             jarray;
    array_comma=          false;  //Dangerous: can mix with prev. array ! Fix it later.
    elems++;
    key_start[elems]=  0;
    key_end[elems]=    0;
    break;
  case '"':
    if (jpair != cur_type)
      parent_types[elems]=  cur_type;
    else
      switch_to_pair= true;
    val_start[elems]=     pos;
    if (jarray == cur_type)
      is_key= false;
    else
      is_key= !is_key;
    cur_type=             jstring;
    array_comma=          false;
    elems++;
    break;
  case ':':
    if (jobject != cur_type && jpair != cur_type)
      return false;
    key_start[elems]=     val_start[elems];
    key_end[elems]=       pos - 1;
    cur_type=             jpair;
    //elems++;
    break;
  case '\n':
  case '\t':
  case ' ' :
    break;
  default:
    return false;
    break;
  }
  return true;
};


bool JParser::parse_object(char c, int pos)
{
  switch (c)
  {
  case '}':
    if (is_key)
      return false;
    elems--;
    int kend;
    int kstart;
    if (0 == (key_end[elems] - key_start[elems]))
    {
      kend= 0;
      kstart= 0;
    }
    else
    {
      kend= key_end[elems] - key_start[elems] + 1;
      kstart= key_start[elems];
    }
    stop_parsing= !this->callback_func(kstart, kend, val_start[elems], 
      pos - val_start[elems] + 1, elems, jarray == parent_types[elems], cur_type);
    cur_type= parent_types[elems];
    waiting_closure= false;
    break;
  default:
    return parse_default(c, pos);
    break;
  }
  return true;
};


bool JParser::parse_array(char c, int pos)
{
  switch (c)
  {
  case ']':
    if (array_comma)
    {
      return false;
    }
    elems--;
    int kend;
    int kstart;
    if (0 == (key_end[elems] - key_start[elems]))
    {
      kend= 0;
      kstart= key_start[elems];
    }
    else
    {
      kend= key_end[elems] - key_start[elems] + 1;
      kstart= key_start[elems];
    }
    stop_parsing= !this->callback_func(kstart, kend, val_start[elems], 
      pos - val_start[elems] + 1, elems, jarray == parent_types[elems], cur_type);
    cur_type= parent_types[elems];
    waiting_closure= false;
    break;
  case ',':
    if (!array_comma)
      array_comma= true;
    else
      return false;
    waiting_closure= false;
    break;
  default:
    return parse_pair_or_array(c, pos);
    break;
  }
  return true;
};


bool JParser::parse_string(char c, int pos)
{
  switch (c)
  {
  case '"':
    //parse quote
    if ('\\' != doc[pos - 1])
    {
      elems--;
      if (is_key)
      {
        key_start[elems]= val_start[elems];
        key_end[elems]=   pos;
      }
      else
      {
        int kend;
        int kstart;
        if (0 == (key_end[elems] - key_start[elems]))
        {
          kend= 0;
          kstart= 0;
        }
        else
        {
          kend= key_end[elems] - key_start[elems] + 1;
          kstart= key_start[elems];
        }
        // remove quotes
        stop_parsing= !this->callback_func(kstart, kend, val_start[elems], pos - val_start[elems] + 1, elems, jarray == parent_types[elems], jstring);
        waiting_closure= true;
      }
      if (switch_to_pair) {
        switch_to_pair= false;
        cur_type= jpair;
      }
      else
        cur_type= parent_types[elems];
    }
    break;
  default:
    //save values in string
    if ('\\' == doc[pos - 1])
    {
      switch (c)
      {
      case '\\':
      case '/':
      case 'b':
      case 'f':
      case 'n':
      case 'r':
      case 't':
        break;
      case 'u':
        /* TODO: fix bug #70570 */
        parent_types[elems]=  cur_type;
        val_start[elems]=     pos;
        cur_type=             jutf;
        elems++;
        break;
      default:
        return false;
      }
    }
    break;
  }
  return true;
};


bool JParser::parse_pair(char c, int pos)
{
  switch (c)
  {
  case ',':
    waiting_closure= false;
    break;
  case ']':
  case '}':
    if (is_key
      || ((jdefault != parent_types[elems]
        && jpair != parent_types[elems])
        &&(('}' == c && jobject != parent_types[elems])
      || (']' == c && jarray != parent_types[elems]))
      ))
    {
      return false;
    }
    elems--;
    if (',' == doc[pos - 1])
      return false;
    int kend;
    int kstart;
    if (0 == (key_end[elems] - key_start[elems]))
    {
      kend= 0;
      kstart= 0;
    }
    else
    {
      kend= key_end[elems] - key_start[elems] + 1;
      kstart= key_start[elems];
    }
    stop_parsing= !this->callback_func(kstart, kend, val_start[elems], 
      pos - val_start[elems] + 1, elems, jarray == parent_types[elems], cur_type);
    cur_type= parent_types[elems];
    waiting_closure= false;
    break;
  default:
    return parse_pair_or_array(c, pos);
    break;
  }
  return true;
};


bool JParser::parse_pair_or_array(char c, int pos)
{
  switch (c)
  {
  case 't':
    parent_types[elems]=  cur_type;
    val_start[elems]=     pos;
    cur_type=             jtrue;
    elems++;
    array_comma=          false;
    break;
  case 'f':
    parent_types[elems]=  cur_type;
    val_start[elems]=     pos;
    cur_type=             jfalse;
    elems++;
    array_comma=          false;
    break;
  case 'n':
    parent_types[elems]=  cur_type;
    val_start[elems]=     pos;
    cur_type=             jnull;
    elems++;
    array_comma=          false;
    break;
  case '-':
  case '0':
  case '1':
  case '2':
  case '3':
  case '4':
  case '5':
  case '6':
  case '7':
  case '8':
  case '9':
    //-?(([1-9][0-9]*)|0)(\.(([0-9]+[e|E][+|-]?)?[0-9]+)?)?
    parent_types[elems]=  cur_type;
    val_start[elems]=     pos;
    cur_type=             jnum;
    elems++;
    array_comma=          false;
    break;
  default:
    return parse_default(c, pos);
    break;
  }
  return true;
};


bool JParser::parse_true(char c, int pos)
{
  switch (pos - val_start[elems - 1])
  {
  case 1:
    if ('r' != c)
      return false;
    break;
  case 2:
    if ('u' != c)
      return false;
    break;
  case 3:
    int kend;
    int kstart;
    
    if (is_key)
      is_key= false;
    
    if ('e' != c)
      return false;
    
    elems--;
    if (0 == (key_end[elems] - key_start[elems]))
    {
      kend= 0;
      kstart= 0;
    }
    else
    {
      kend= key_end[elems] - key_start[elems] + 1;
      kstart= key_start[elems];
    }
    stop_parsing= !this->callback_func(kstart, kend, val_start[elems], 
      pos - val_start[elems] + 1, elems, jarray == parent_types[elems], cur_type);
    cur_type= parent_types[elems];
    waiting_closure= true;
    break;
  default:
    return false;
  }
  return true;
};


bool JParser::parse_false(char c, int pos)
{
  switch (pos - val_start[elems - 1])
  {
  case 1:
    if ('a' != c)
      return false;
    break;
  case 2:
    if ('l' != c)
      return false;
    break;
  case 3:
    if ('s' != c)
      return false;
    break;
  case 4:
    int kend;
    int kstart;
    
    if (is_key)
      is_key= false;
    
    if ('e' != c)
      return false;
    
    elems--;
    if (0 == (key_end[elems] - key_start[elems]))
    {
      kend= 0;
      kstart= 0;
    }
    else
    {
      kend= key_end[elems] - key_start[elems] + 1;
      kstart= key_start[elems];
    }
    stop_parsing= !this->callback_func(kstart, kend, val_start[elems], 
      pos - val_start[elems] + 1, elems, jarray == parent_types[elems], cur_type);
    cur_type= parent_types[elems];
    waiting_closure= true;
    break;
  default:
    return false;
  }
  
  return true;
};


bool JParser::parse_null(char c, int pos)
{
  switch (pos - val_start[elems - 1])
  {
  case 1:
    if ('u' != c)
      return false;
    break;
  case 2:
    if ('l' != c)
      return false;
    break;
  case 3:
    int kend;
    int kstart;
    
    if (is_key)
      is_key= false;
    
    if ('l' != c)
      return false;
    
    elems--;
    if (0 == (key_end[elems] - key_start[elems]))
    {
      kend= 0;
      kstart= 0;
    }
    else
    {
      kend= key_end[elems] - key_start[elems] + 1;
      kstart= key_start[elems];
    }
    stop_parsing= !this->callback_func(kstart, kend, val_start[elems], 
      pos - val_start[elems] + 1, elems, jarray == parent_types[elems], cur_type);
    cur_type= parent_types[elems];
    waiting_closure= true;
    break;
  default:
    return false;
  }
  return true;
};


bool JParser::parse_num(char c, int pos)
{
  //-?(([1-9][0-9]*)|0)(\.(([0-9]+[e|E][+|-]?)?[0-9]+)?)?
  switch (c)
  {
  case '.':
  case '-':
  case '+':
  case 'e':
  case 'E':
  case '0':
  case '1':
  case '2':
  case '3':
  case '4':
  case '5':
  case '6':
  case '7':
  case '8':
  case '9':
    return true;
  default:
    //store num
    pcre* regex;
    int reti;
    const char *re_error;
    int re_erroroffset;
    int re_vector[30];
    char* num;
    int kend;
    int kstart;
    
    if (is_key)
      is_key= false;
    
    elems--;
    
    num= new char[pos - val_start[elems] + 1];
    
    memcpy(num, &doc[val_start[elems]], pos - val_start[elems]);
    num[pos - val_start[elems]]= '\0';
    
    regex= pcre_compile("^-?(([1-9][0-9]*)|0)(\\.?(([0-9]*[e|E][+|-]?)?[0-9]+)?)?$", PCRE_EXTENDED, &re_error, &re_erroroffset, 0);
    if (NULL == regex)
    {
      delete[] num;
      //TODO: write error
      return false;
    }
    reti= pcre_exec(regex, 0, num, strlen(num), 0, 0, re_vector, 30);
    delete[] num;
    if (0 > reti)
    {
      //TODO: check if( reti == REG_NOMATCH )
      //and return appropriate error message
      pcre_free(regex);
      return false;
    }
    pcre_free(regex);
    if (0 == (key_end[elems] - key_start[elems]))
    {
      kend= 0;
      kstart= 0;
    }
    else
    {
      kend= key_end[elems] - key_start[elems] + 1;
      kstart= key_start[elems];
    }
    stop_parsing= !this->callback_func(kstart, kend, val_start[elems], pos - val_start[elems], elems, jarray == parent_types[elems], cur_type);
    cur_type= parent_types[elems];
    waiting_closure= true;
    return parse_char(c, pos);
  }
};


bool JParser::parse_utf_escape(char c, int pos)
{
  switch (c)
  {
  case '0':
  case '1':
  case '2':
  case '3':
  case '4':
  case '5':
  case '6':
  case '7':
  case '8':
  case '9':
  case 'a':
  case 'A':
  case 'b':
  case 'B':
  case 'c':
  case 'C':
  case 'd':
  case 'D':
  case 'e':
  case 'E':
  case 'f':
  case 'F':
    break;
  default:
    return false;
  }
  if (4 == (pos - val_start[elems - 1]))
  {
    elems--;
    cur_type= parent_types[elems];
    return true;
  }
  else if ( 4 < (pos - val_start[elems - 1]))
    return false;
  else
    return true;
};


bool JParser::callback_func(int key_start, int key_len, int val_start, int val_len, int pos, bool is_array_elem, int type_flag)
{
  cleanup_key_map();
  
  if (jobject == parent_types[elems])
  {
    if (!this->store_key(key_start, key_len))
    {
      this->parse_error= true;
      return true;
    }
  }
  
  return callback->callback_func(key_start, key_len, val_start, val_len, pos, is_array_elem, type_flag);
};


bool JParser::store_key(int key_start, int key_len)
{
  bool is_different= false;
  unsigned int iterations= 0;
  
  cleanup_key_map();
  
  
  std::pair <std::map<int, std::pair<int, int> >::iterator, std::map<int, std::pair<int, int> >::iterator> et = elem_keys.equal_range(elems);
  
  for (std::map<int, std::pair<int, int> >::iterator it = et.first; it != et.second; it++, iterations++)
  {
    for (unsigned int i= 0; i < key_len; i++)
      if (i < (it->second).second)
      {
        if (doc[key_start + i] != doc[(it->second).first + i])
        {
          is_different= true;
          break;
        }
      }
      else
      {
        is_different= true;
        break;
      }
  }
  
  
  if (is_different || 0 == iterations)
  {
    this->elem_keys.insert(std::pair<int, std::pair<int, int> >(elems, std::pair<int,int>(key_start, key_len)));
    
    return true;
  }
  
  return false;
};


void JParser::cleanup_key_map()
{
  int elem= elems + 1;
  std::map<int,std::pair<int, int> >::iterator it;
  while ((it = elem_keys.find(elem++)) != elem_keys.end())
    elem_keys.erase(it);
};

/* :tabSize=2:indentSize=2:noTabs=true:folding=custom:collapseFolds=1: */
