/* 
  Copyright (c) 2013 - 2014, Oracle and/or its affiliates. All rights reserved.
  
  This program is free software; you can redistribute it and/or modify it under 
  the terms of the GNU General Public License as published by the Free Software 
  Foundation; version 2 of the License.
  
  This program is distributed in the hope that it will be useful, but WITHOUT ANY 
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS 
  FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License along with this 
  program; if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth 
  Floor, Boston, MA 02110-1301  USA
*/
#pragma once

#ifndef JPARSER
#define JPARSER

#include "jstack.h"
#include "JCallback.h"
#include "jutil.h"

#include <string.h>
#include <stdio.h>
#include <ctype.h>

/* Workaround for bug #28184 which is not in 5.5 */
#undef min
#undef max
#include <map>
#ifndef WITH_BUNDLED_PCRE
#include <pcre.h>
#else
#include "pcre.h"
#endif


/* 
  JSON parser class
*/

class JParser
{
private:
  JCallback* callback;                            /* callback function */
  char* doc;                                      /* document to be parsed */
  enum_jtype cur_type;                            /* current element type */
  int elems;                                      /* number of elems proceed */
  char* buffer;                                   /* buffer to hold parsed elements */
  bool stop_parsing;                              /* do we need to proceed more results? */
  enum_jtype* parent_types;                       /* array to store parent types */
  int* key_start;                                 /* Current key start position */
  int* key_end;                                   /* Current key end position */
  int* val_start;                                 /* Current value start position */
  int* val_end;                                   /* Current value end position */
  bool array_comma;                               /* Parsing array element */
  bool is_key;                                    /* Is sting a key? */
  bool switch_to_pair;                            /* Is a string a part of a pair? */
  bool parse_error;                               /* If parse error occurred during parsing */
  bool waiting_closure;                           /* Element proceed, expecting closing character */
  std::map<int, std::pair<int, int> > elem_keys;  /* Map of keys by hierarchy,  */
                                                  /* so we can check for duplicates */
  
public:
  
  /* 
    Constructor.
    
    @param[in]  document            document to parse
    @param[in]  callback_function   callback function
  */
  
  JParser(char* document, JCallback* callback_function);
  ~JParser();
  
  
  /* 
    Parses document function
    Splits string into chars, then
    calls parse_char
    
    @return Operation status
      @retval true  OK
      @retval false Failure
  */
  
  bool parse();
  
  
private:
  /* 
    Parses char function
    Calls parse_default and its companion
    depending from context
    
    @param[in]  c   char to parse
    
    @return Operation status
      @retval true  OK
      @retval false Failure
  */
  
  bool parse_char(char c, int pos);
  
  
  /* 
    Functions, which called depending
    from actual parse state.
    Do real job.
    
    @param[in]  c   char to parse
    
    @return Operation status
      @retval true  OK
      @retval false Failure
  */
  bool parse_default(char c, int pos);
  bool parse_object(char c, int pos);
  bool parse_array(char c, int pos);
  bool parse_string(char c, int pos);
  bool parse_pair(char c, int pos);
  bool parse_pair_or_array(char c, int pos);
  bool parse_true(char c, int pos);
  bool parse_false(char c, int pos);
  bool parse_null(char c, int pos);
  bool parse_num(char c, int pos);
  bool parse_utf_escape(char c, int pos);
  
  
  /* 
    Checks if key in the object is unique
    then calls callback->callback_func
    
    @param[in]  key_start       start position of the key
    @param[in]  key_len         length of the key
    @param[in]  val_start       start position of the value
    @param[in]  val_len         length of the value
    @param[in]  pos             key position in the tree
    @param[in]  is_array_elem   is this element of an array?
    @param[in]  type_flag       enum_jtype of the element
    
    @return Operation status
      @retval true  OK
      @retval false Failure
  */
  bool callback_func(int key_start, int key_len, int val_start, int val_len, int pos, bool is_array_elem = false, int type_flag= jdefault);
  
  
  /* 
    Stores keys temporarily in elem_keys.
    
    @param[in]  key_start       start position of the key
    @param[in]  key_len         length of the key
    
    @return Operation status
      @retval true  key is unique, it is OK to continue parsing
      @retval false same key is already exists, stop parsing:
                    JSON document is invalid
  */
  bool store_key(int key_start, int key_len);
  
  
  /* 
    Cleanups elem_keys map:
    erases all maps with key, bigger than elems
  */
  void cleanup_key_map();
  
};

#endif /* JPARSER */

/* :tabSize=2:indentSize=2:noTabs=true:folding=custom:collapseFolds=1: */
