/* 
  Copyright (c) 2013 - 2014, Oracle and/or its affiliates. All rights reserved.
  
  This program is free software; you can redistribute it and/or modify it under 
  the terms of the GNU General Public License as published by the Free Software 
  Foundation; version 2 of the License.
  
  This program is distributed in the hope that it will be useful, but WITHOUT ANY 
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS 
  FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License along with this 
  program; if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth 
  Floor, Boston, MA 02110-1301  USA
*/
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "my_json.h"


bool json_valid(char* doc)
{
  bool result;
  
	JCallback_valid* jcv= new JCallback_valid();
	JParser* jp= new JParser(doc, jcv);
	result= jp->parse();
	
	delete jp;
	delete jcv;
	
	return result;
}


char* json_extract(char* doc, JKey** keys, int klen)
{
  char* retval;
  JCallback_extract* je= new JCallback_extract(doc, keys, klen);
  JParser* jp= new JParser(doc, je);
  bool result= jp->parse();
  
  if (!result)
  {
    delete jp;
    delete je;
    return NULL;
  }
  
  retval = je->extract();
  
  delete jp;
  delete je;
  
  return retval;
}


char* json_replace(char* doc, JKey** keys, int klen, char* new_value)
{
  char* retval;
  int new_value_size= strlen(new_value);
  char* test_validity= new char[new_value_size + 3];
  
  test_validity[0]= '[';
  memcpy(&test_validity[1], new_value, new_value_size);
  test_validity[new_value_size + 1]= ']';
  test_validity[new_value_size + 2]= '\0';
  
  if (!json_valid(test_validity))
  {
    jdump(JSON_ERROR, JSON_FUNCTION_ERROR, "New value is not valid JSON document");
    delete[] test_validity;
    return NULL;
  }
  
  delete[] test_validity;
  
  JCallback_replace* jr= new JCallback_replace(doc, keys, klen, new_value);
  JParser* jp= new JParser(doc, jr);
  bool result= jp->parse();
  
  if (!result)
  {
    delete jp;
    delete jr;
    return NULL;
  }
  
  retval = jr->get_doc();
  
  delete jp;
  delete jr;
  
  return retval;
}


char* json_append(char* doc, JKey** keys, int klen, char* value)
{
  char* retval;
  int value_size= strlen(value);
  char* test_validity= new char[value_size + 3];
  
  test_validity[0]= '[';
  memcpy(&test_validity[1], value, value_size);
  test_validity[value_size + 1]= ']';
  test_validity[value_size + 2]= '\0';
  
  if (!json_valid(test_validity))
  {
    jdump(JSON_ERROR, JSON_FUNCTION_ERROR, "New value is not valid JSON document");
    delete[] test_validity;
    return NULL;
  }
  
  delete[] test_validity;
  
  JCallback_append* ja= new JCallback_append(doc, keys, klen, value);
  JParser* jp= new JParser(doc, ja);
  bool result= jp->parse();
  
  if (!result)
  {
    delete jp;
    delete ja;
    return NULL;
  }
  
  retval = ja->get_doc();
  
  delete jp;
  delete ja;
  
  return retval;
}


char* json_remove(char* doc, JKey** keys, int klen)
{
  char* retval;
  JCallback_remove* jr= new JCallback_remove(doc, keys, klen);
  JParser* jp= new JParser(doc, jr);
  bool result= jp->parse();
  
  if (!result)
  {
    delete jp;
    delete jr;
    return NULL;
  }
  
  retval = jr->get_doc();
  
  delete jp;
  delete jr;
  
  return retval;
}


char* json_set(char* doc, JKey** keys, int klen, char* new_value)
{
  char* retval;
  int new_value_size= strlen(new_value);
  char* test_validity= new char[new_value_size + 3];
  
  test_validity[0]= '[';
  memcpy(&test_validity[1], new_value, new_value_size);
  test_validity[new_value_size + 1]= ']';
  test_validity[new_value_size + 2]= '\0';
  
  if (!json_valid(test_validity))
  {
    jdump(JSON_ERROR, JSON_FUNCTION_ERROR, "New value is not valid JSON document");
    delete[] test_validity;
    return NULL;
  }
  
  delete[] test_validity;
  
  JCallback_set* js= new JCallback_set(doc, keys, klen, new_value);
  JParser* jp= new JParser(doc, js);
  bool result= jp->parse();
  
  if (!result)
  {
    delete jp;
    delete js;
    return NULL;
  }
  
  retval = js->get_doc();
  
  delete jp;
  delete js;
  
  return retval;
}


char* json_merge(char** documents, int dlen, bool check_for_validity, bool deep)
{
  char* retval;
  char first_bracket, last_bracket;
  int total_len= 1;
  int actual_len= 0;
  
  if (!(first_bracket= check_first_bracket(documents[0])) ||
    (check_for_validity && !json_valid(documents[0])))
  {
    jdump(JSON_ERROR, JSON_FUNCTION_ERROR, "Error processing first document in JSON_MERGE");
    return NULL;
  }
    
  if (OPENING_CURLY_BRACKET == first_bracket)
    last_bracket= CLOSING_CURLY_BRACKET;
  else if (OPENING_SQUARE_BRACKET == first_bracket)
    last_bracket= CLOSING_SQUARE_BRACKET;
  else
  {
    jdump(JSON_ERROR, JSON_FUNCTION_ERROR, "No bracket in the first document, nothing to merge");
    return NULL; /* no bracket */
  }
  
  if (deep && OPENING_CURLY_BRACKET == first_bracket)
  {
    JCallback_merge** jms= new JCallback_merge*[dlen];
    Jtree_elem* jte;
    Jtree_elem* first;
    char* result;
    int jms_real_len= 0;
    int root_doc_ind= 0;
    
    for (int i= 0; i < dlen; i++)
    {
      if (!check_first_bracket(documents[i], first_bracket))
      {
        jdump(JSON_ERROR, JSON_FUNCTION_ERROR, "Open bracket does not match or document %d is invalid", i + 1);
        break;
      }
      
      jms[i]= new JCallback_merge(documents[i]);
      JParser* jp= new JParser(documents[i], jms[i]);
      bool parse_result= jp->parse();
      delete jp;
      
      if (parse_result)
      {
        stack_jelem* st= jms[i]->get_stack();
        Jtree_elem* root= jms[i]->get_merge_tree();
        if (0 == i)
          jte= root;
        else if (0 == jte->num_childs)
        {
          jte= root;
          root_doc_ind= i;
        }
        else
          merge_tree(jte, root);
      }
      else
      {
        break;
      }
      
      jms_real_len++;
    }
    
    if (0 < jte->val_offset)
      result= new char[strlen(documents[root_doc_ind]) + jte->val_offset + 2 * dlen];
    else
      result= new char[strlen(documents[root_doc_ind]) + 2 * dlen];
    
    for (int i= 0; i < strlen(documents[root_doc_ind]); i++)
    {
      result[i]= documents[root_doc_ind][i];
    }
    if (0 < jte->val_offset)
      for (int i= strlen(documents[root_doc_ind]); i < strlen(documents[root_doc_ind]) + jte->val_offset + 2 * dlen; i++)
        result[i]= '\0';
    else
      for (int i= strlen(documents[root_doc_ind]); i < strlen(documents[root_doc_ind]) + 2 * dlen; i++)
        result[i]= '\0';
      
    fill_result(jte, result);
    for (int i= 0; i < jms_real_len; i++)
    {
      delete jms[i];
    }
    delete[] jms;
    return result;
  }
  else
  {
    for (int i= 0; i < dlen; i++)
      total_len+= strlen(documents[i]);
  
    retval= new char[strlen(documents[0]) + total_len];
  
    memcpy(&retval[0], &(documents[0])[0], strlen(documents[0]));
    actual_len+= strlen(documents[0]);
    
    for (int i= 1; i < dlen; i++)
    {
      int first_bracket_pos= 0;
      int last_bracket_pos= actual_len - 1;
      bool valid= true;
      bool stop_loop= false;
      
      while (0 < last_bracket_pos)
      {
        switch(retval[last_bracket_pos])
        {
        case CLOSING_CURLY_BRACKET:
        case CLOSING_SQUARE_BRACKET:
          if (last_bracket != retval[last_bracket_pos])
            valid= false;
          stop_loop= true;
          break;
        case ' ':
        case '\n':
        case '\r':
        case '\t':
          break;
        default:
          valid= false;
          break;
        }
        if (stop_loop)
          break;
        last_bracket_pos--;
      }
      
      if (!valid || 0 == last_bracket_pos)
      {
        jdump(JSON_ERROR, JSON_FUNCTION_ERROR, "Closing bracket does not match in the document %d", i + 1);
        break;
      }
      
      stop_loop= false;
      
      if (!check_first_bracket(documents[i], first_bracket) ||
      (check_for_validity && !json_valid(documents[i])))
      {
        jdump(JSON_ERROR, JSON_FUNCTION_ERROR, "Open bracket does not match or document %d is invalid", i + 1);
        break;
      }
      
      retval[last_bracket_pos]= ',';
      retval[last_bracket_pos + 1]= ' ';
      
      memcpy(&retval[last_bracket_pos + 2], &(documents[i])[first_bracket_pos + 1], strlen(documents[i]) - first_bracket_pos - 1);
      actual_len= last_bracket_pos + 2 + strlen(documents[i]) - first_bracket_pos - 1;
    }
    
    retval[actual_len]= '\0';
  }
  
  return retval;
}


/* 
  Search whole doc
  if not found - return NULL
  if found - record last pos
  parse until last pos in value
  check value
  if exists - return
  if not - check earlier values
*/
char* json_search(char* doc, char* value, int search_pos, bool path_format)
{
  char* retval;
  JCallback_search* js= new JCallback_search(doc, value, search_pos);
  JParser* jp= new JParser(doc, js);
  bool result= jp->parse();
  
  if (!result)
  {
    delete jp;
    delete js;
    return NULL;
  }
  
  retval= js->get_path(path_format);
  
  jdump(JSON_NOTE, JSON_FUNCTION_ERROR, "retval: %s", retval);
  delete jp;
  jdump(JSON_NOTE, JSON_FUNCTION_ERROR, "after delete jp");
  delete js;
  jdump(JSON_NOTE, JSON_FUNCTION_ERROR, "after delete js");
  
  return retval;
}


bool json_contains_key(char* doc, JKey** keys, int klen)
{
  JCallback_contains_key* jck= new JCallback_contains_key(doc, keys, klen);
  JParser* jp= new JParser(doc, jck);
  bool result= jp->parse();
  
  if (!result)
  {
    delete jp;
    delete jck;
    return false;
  }
  
  result = jck->contains();
  
  delete jp;
  delete jck;
  
	return result;
}


char* json_test_parser(char* doc)
{
  JCallback_test_parser* jtp= new JCallback_test_parser();
  JParser* jp= new JParser(doc, jtp);
  bool result= jp->parse();
  
  char* res= jtp->get_parse_tree(doc);
  
  delete jp;
  delete jtp;
  
	return res;
}


int json_depth(char* doc)
{
  bool parse_result;
  int result;
  
	JCallback_depth* jcd= new JCallback_depth();
	JParser* jp= new JParser(doc, jcd);
	parse_result= jp->parse();
	
  if (!parse_result)
  {
    delete jp;
    delete jcd;
    return 0;
  }
  
  result= jcd->get_depth();
  
  delete jp;
  delete jcd;
  
	return result;
}


int json_count(char* doc, JKey** keys, int klen)
{
  bool parse_result;
  int result;
  JParser* jp;
	
  JCallback_count* jcc= new JCallback_count();
	
  if (0 < klen)
  {
    char* temp_doc= json_extract(doc, keys, klen);
    
    if (!temp_doc)
    {
      delete jcc;
      jdump(JSON_NOTE, JSON_FUNCTION_ERROR, "Cannot find element");
      return 0;
    }
    
    jp= new JParser(temp_doc, jcc);
    parse_result= jp->parse();
    delete[] temp_doc;
  }
  else
  {
    jp= new JParser(doc, jcc);
    parse_result= jp->parse();
  }
	
  if (!parse_result)
  {
    delete jp;
    delete jcc;
    return 0;
  }
  
  result= jcc->get_count();
  
  delete jp;
  delete jcc;
  
	return result;
}


char* json_version()
{
  char* version= new char[27];
  sprintf(version, "MySQL JSON UDFs %d.%d.%d-labs", VERSION_MAJOR, VERSION_MINOR, VERSION_PATCH);

  return version;
}


char check_first_bracket(char* doc, char bracket)
{
  int first_bracket= 0;
  char found= '\0';
  bool stop_loop= false;
  
  while (strlen(doc) > first_bracket)
  {
    switch(doc[first_bracket])
    {
    case '{':
    case '[':
      stop_loop= true;
      if (!bracket || bracket == doc[first_bracket])
        found= doc[first_bracket];
      break;
    case ' ':
    case '\n':
    case '\r':
    case '\t':
      break;
    default:
      stop_loop= true;
      break;
    }
    if (stop_loop)
      break;
    first_bracket++;
  }
  
  return found;
};


void print_merge_tree(Jtree_elem* root)
{
    char* key= new char[root->st[root->pos_in_stack].key_len + 1];
    char* value= new char[root->st[root->pos_in_stack].val_len + 1];
    for (int i= 0; i < root->st[root->pos_in_stack].key_len; i++)
      key[i]= root->doc[root->st[root->pos_in_stack].key_start + i];
    for (int i= 0; i < root->st[root->pos_in_stack].val_len; i++)
      value[i]= root->doc[root->st[root->pos_in_stack].val_start + i];
    key[root->st[root->pos_in_stack].key_len]= '\0';
    value[root->st[root->pos_in_stack].val_len]= '\0';
    jdump(JSON_NOTE, JSON_FUNCTION_ERROR, "start_pos: %d, key_len: %d, val_start: %d, val_len: %d, pos: %d, obj: %d, is_new: %d", 
      root->st[root->pos_in_stack].key_start, root->st[root->pos_in_stack].key_len, root->st[root->pos_in_stack].val_start, 
      root->st[root->pos_in_stack].val_len, root->pos_in_stack, root->st[root->pos_in_stack].type_flag, root->is_new);
    jdump(JSON_NOTE, JSON_FUNCTION_ERROR, "key: %s, value: %s, offset: %d", key, value, root->val_offset);
  
  if (root->childs)
  {
    for (int i= 0; i < root->num_childs; i++)
    {
      print_merge_tree(root->childs[i]);
    }
  }
};

int fill_result(Jtree_elem* jte, char* result, int total_offset)
{
  if (jte->is_new)
  {
    int elem_len= jte->st[jte->pos_in_stack].val_start + jte->st[jte->pos_in_stack].val_len - jte->st[jte->pos_in_stack].key_start;
    
    int start_pos= jte->val_offset + total_offset - 2 + 1;
    
    for (int i= strlen(result); i >= start_pos; i--)
    {
      result[i + elem_len + 2]= result[i];
    }
    
    result[start_pos]= ',';
    result[start_pos + 1]= ' ';
    
    for (int i= 0; i < elem_len; i++)
    {
      result[start_pos + 2 + i]= jte->doc[jte->st[jte->pos_in_stack].key_start + i];
    }
    
    total_offset+= elem_len + 2;
  }
  else if (0 < jte->num_childs)
  {
    if (jarray == jte->st[jte->pos_in_stack].type_flag)
    {
      int start_pos= jte->st[jte->pos_in_stack].val_start + jte->st[jte->pos_in_stack].val_len + total_offset - 1;
      for (int i= 0; i < jte->num_childs; i++)
      {
        if (0 < jte->childs[i]->val_offset)
        {
          for (int j= strlen(result); j >= start_pos; j--)
          {
            result[j + jte->childs[i]->st[jte->childs[i]->pos_in_stack].val_len + 2]= result[j];
          }
          
          result[start_pos]= ',';
          result[start_pos + 1]= ' ';
          
          for (int j= 0; j < jte->childs[i]->st[jte->childs[i]->pos_in_stack].val_len; j++)
          {
            result[start_pos + 2 + j]= jte->childs[i]->doc[jte->childs[i]->st[jte->childs[i]->pos_in_stack].val_start + j];
          }
          
          total_offset+= jte->childs[i]->st[jte->childs[i]->pos_in_stack].val_len + 2;
          start_pos+= jte->childs[i]->st[jte->childs[i]->pos_in_stack].val_len + 2;
        }
      }
    }
    else
    {
      for (int i= 0; i < jte->num_childs; i++)
      {
        total_offset= fill_result(jte->childs[i], result, total_offset);
      }
    }
  }
  else if (jte->changed)
  {
    if (0 < jte->val_offset)
    {
      for (int i= strlen(result); i > jte->val_start; i--)
      {
        result[i + total_offset + jte->val_offset]= result[i + total_offset];
      }
    }
    else if (0 > jte->val_offset)
    {
      int fin_len= strlen(result);
      for (int i= jte->/* st[jte->pos_in_stack]. */val_start + jte->st[jte->pos_in_stack].val_len + total_offset - jte->val_offset; i < fin_len - jte->val_offset; i++)
      {
        result[i + jte->val_offset]= result[i];
      }
    }
    
    for (int i= 0; i < jte->st[jte->pos_in_stack].val_len; i++)
    {
      result[jte->val_start + total_offset + i]= jte->doc[jte->st[jte->pos_in_stack].val_start + i];
    }
    
    total_offset+= jte->val_offset;
  }
  else
  {
    jdump(JSON_NOTE, JSON_FUNCTION_ERROR, "Untested branch in fill_result");
  }
  return total_offset;
};

void merge_tree(Jtree_elem* jte, Jtree_elem* root)
{
  if (root->childs)
  {
    if (jpair == root->st[root->pos_in_stack].type_flag)
    {
      for (int i = 0; i < root->num_childs; i++)
      {
        int child;
        if (0 <= (child= merge_key_exists(jte, root->childs[i])))
        {
          if (0 == jte->childs[child]->num_childs)
          {
            int val_offset= jte->childs[child]->val_offset+ root->childs[i]->st[root->childs[i]->pos_in_stack].val_len - jte->childs[child]->st[jte->childs[child]->pos_in_stack].val_len;
            int offset= jte->childs[child]->offset + jte->childs[child]->st[jte->childs[child]->pos_in_stack].val_start - root->childs[i]->st[root->childs[i]->pos_in_stack].val_start;
            int val_start= jte->childs[child]->st[jte->childs[child]->pos_in_stack].val_start;
            bool is_new= jte->childs[child]->is_new;
            jte->childs[child]= root->childs[i];
            jte->childs[child]->changed= true;
            jte->childs[child]->val_offset= val_offset;
            jte->childs[child]->offset= offset;
            jte->childs[child]->val_start= val_start;
            jte->childs[child]->is_new= is_new;
          }
          else
          {
            merge_tree(jte->childs[child], root->childs[i]);
          }
          jte->val_offset= jte->val_offset + jte->childs[child]->val_offset;
        }
        else
        {
          jte->childs[jte->num_childs]= root->childs[i];
          jte->childs[jte->num_childs]->is_new= true;
          jte->childs[jte->num_childs]->val_offset= jte->childs[jte->num_childs]->val_offset + jte->st[jte->pos_in_stack].val_start + jte->st[jte->pos_in_stack].val_len;
          jte->changed= true;
          jte->val_offset= jte->val_offset + root->st[root->childs[i]->pos_in_stack].val_len 
          + (root->st[root->childs[i]->pos_in_stack].val_start - root->st[root->childs[i]->pos_in_stack].key_start) + 2
          ;
          jte->num_childs++;
        }
      }
    }
    else
    {
      for (int i = 0; i < root->num_childs; i++)
      {
        jte->childs[jte->num_childs]= root->childs[i];
        jte->childs[jte->num_childs]->is_new= true;
        jte->childs[jte->num_childs]->val_offset= jte->childs[jte->num_childs]->val_offset + 2;
        jte->changed= true;
        jte->val_offset= jte->val_offset + root->st[root->childs[i]->pos_in_stack].val_len 
        + 2
        ;
        jte->num_childs++;
      }
    }
  }
  else if (jpair == root->st[root->pos_in_stack].type_flag)
  {
    if (merge_key_exists(jte, root))
    {
      jte= root;
      return merge_tree(jte, root);
    }
  }
  else
  {
    jdump(JSON_NOTE, JSON_FUNCTION_ERROR, "untested branch");
  }
};


int merge_key_exists(Jtree_elem* root, Jtree_elem* elem)
{
  for (int i= 0; i < root->num_childs; i++)
  {
    bool found= true;
    Jtree_elem* rchild= root->childs[i];
    
    if (rchild->st[rchild->pos_in_stack].key_len != elem->st[elem->pos_in_stack].key_len)
      continue;
    
    for (int j= 0; j < rchild->st[rchild->pos_in_stack].key_len; j++)
    {
      if (rchild->doc[rchild->st[rchild->pos_in_stack].key_start + j]
        != elem->doc[elem->st[elem->pos_in_stack].key_start + j])
      {
        found= false;
        break;
      }
    }
    
    if (!found)
    {
      continue;
    }
    else
    {
      return i;
    }
  }
  
  return -1;
};

/*
* :tabSize=2:indentSize=2:noTabs=true:
* :folding=custom:collapseFolds=1:
*/
