/* 
  Copyright (c) 2013 - 2014, Oracle and/or its affiliates. All rights reserved.
  
  This program is free software; you can redistribute it and/or modify it under 
  the terms of the GNU General Public License as published by the Free Software 
  Foundation; version 2 of the License.
  
  This program is distributed in the hope that it will be useful, but WITHOUT ANY 
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS 
  FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License along with this 
  program; if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth 
  Floor, Boston, MA 02110-1301  USA
*/
#pragma once

#ifndef MY_JSON
#define MY_JSON

#include "JParser.h"
#include "jstack.h"

#define OPENING_CURLY_BRACKET '{'
#define CLOSING_CURLY_BRACKET '}'
#define OPENING_SQUARE_BRACKET '['
#define CLOSING_SQUARE_BRACKET ']'


/* Tree of all merged documents */
class JMerge_tree_elem: public Jtree_elem
{
public:
  stack_jelem* st;            /* Pointer to stack_jelem */
  char* doc;                  /* Pointer to doc */
  int new_val_len;            /* New length of value if changed */
  bool changed;               /* Was value changed or not? */
  int offset;                 /* Offset from key_start and val_start positions */
};                                                                                      


/*
  Checks if doc is valid JSON document.
  
  SYNOPSIS
    json_valid()
    doc      JSON document
  
  RETURN VALUES
    true  ok
    false  document is invalid
*/
bool json_valid(char* doc);


/*
  Extracts value of the specified key.
  
  SYNOPSIS
    json_extract()
    doc     JSON document
    keys    key to search
    klen    lenfth of key array
    
  RETURN VALUES
    char*  value of the key.
    NULL  if no key found
    NULL  if document is invalid warning generated

*/

char* json_extract(char* doc, JKey** keys, int klen);


/*
  Updates value of specified key.
  
  SYNOPSIS
    json_replace()
    doc        JSON document
    keys       key which value should be replaced
    klen       length of key array
    new_value  new value of the key
    
  RETURN VALUES
    char*  document with replaced key
    char*  original document if no such an element found
    NULL   if document is invalid

*/

char* json_replace(char* doc, JKey** keys, int klen, char* new_value);

/*
  Inserts new element.
  
  SYNOPSIS
    json_append()
    doc        JSON document
    key        key for new element
    klen       length of key array
    value      value of the new element

    
  RETURN VALUES
    char*  document with appended element
    NULL   if document is invalid
    char*  If key or value is invalid original document is returned and a warning is generated.

*/

char* json_append(char* doc, JKey** keys, int klen, char* value);

/*
  Removes element specified by the key.
  
  SYNOPSIS
    doc    JSON document
    keys   key of the element to be removed
    klen   length of key array

    
  RETURN VALUES
    char*  document without the element
    NULL   if document is invalid

*/

char* json_remove(char* doc, JKey** keys, int klen);

/*
  Performs kind of INSERT ... ON DUPLICATE KEY UPDATE operation.
  
  SYNOPSIS
    doc    JSON document
    keys   key of the element
    klen   length of key array
    value  new value of the element

    
  RETURN VALUES
    char*  document with updated or inserted element
    NULL   if document is invalid

*/

char* json_set(char* doc, JKey** keys, int klen, char* new_value);

/*
  Merges 2 or more documents into one.
  
  SYNOPSIS
    docs  				JSON documents to merge
    dlen  				length of doc array
    check_for_validity	should this function check for validity or not

    
  RETURN VALUES
    char*  first document with following documents appended
        If one of following documents does not contain opening or closing curly bracket 
        or check_for_validity is true and document is not valid JSON
        first valid documents merged are returned
    NULL  if first document is invalid

*/

char* json_merge(char** documents, int dlen, bool check_for_validity= false, bool deep= false);


/*
  Searches for specified value in the document.
  
  SYNOPSIS
    json_search()
    doc         JSON document
    value 	    value to search
    search_pos	which occurrence to search?
    path_format if path format should be used ($.key[arr_elem])
    
    
  RETURN VALUES
    char*  key path of the element which contains the value in reverse order
    NULL   if document is invalid.

*/

char* json_search(char* doc, char* value, int search_pos= 1, bool path_format= false);

/*
  Checks if documents contains specified key.
  
  SYNOPSIS
    json_contains_key()
    doc    JSON document
    keys   key to search for
    klen   length of key array
    
    
  RETURN VALUES
    TRUE   if key exists
    FALSE  if key does not exist
    NULL   if document is invalid

*/

bool json_contains_key(char* doc, JKey** keys, int klen);


/*
  Parses document and returns parse tree.
  Designed for test purpose only
  
  SYNOPSIS
    json_test_parser()
    doc    JSON document
    
    
  RETURN VALUES
    char*  parse tree if document is valid
    NULL  if document is invalid

*/

char* json_test_parser(char* doc);



/*
  Counts depth of the JSON document.
  
  Thus, for document {"a": {"b": "c", "d": "e", "f": "g"}} result will be 4
  And for document {"a": "b", "c": "d", "e": "f", "g": "h"} result will be 2
  
  SYNOPSIS
    json_depth()
    doc      JSON document
  
  RETURN VALUES
    long long  depth of the document
    NULL       document is invalid
*/
int json_depth(char* doc);



/*
  Counts number of childs of an element of the JSON document.
  
  Thus, for document {"a": {"b": "c", "d": "e", "f": "g"}} result will be 1
  And for document {"a": "b", "c": "d", "e": "f", "g": "h"} result will be 4
  
  SYNOPSIS
    json_count()
    doc      JSON document
    keys   	 key to search for
    klen     length of key array
  
  RETURN VALUES
    long long  number of childs of the root element
    NULL       document is invalid
*/
int json_count(char* doc, JKey** keys, int klen);


/* 
  Returns version of JSON functions
  
  SYNOPSIS
    json_version()
    
  RETURN VALUES
    char*   version of JSON functions
*/
char* json_version();

/* 
Checks if document contains opening curly or square bracket bracket

SYNOPSIS
  check_first_bracket()
  doc    	JSON document
  bracket	if we search for a specific bracket
  
RETURN VALUES
  bracket   if contains opening bracket
  NULL   	  if misses the bracket
*/
char check_first_bracket(char* doc, char bracket= '\0');


/* 
Prints tree of merged documents

This is helper function which should be used for debugging purpose only

SYNOPSIS
  print_merge_tree();
  root        tree root
  
RETURN VALUES
  void
*/
void print_merge_tree(Jtree_elem* root);


/* 
Merges individual trees to one

SYNOPSIS
  merge_tree();
  jte         main tree root
  root        subsequent tree root
  
RETURN VALUES
  void
*/
void merge_tree(Jtree_elem* jte, Jtree_elem* root);


/* 
Checks if key exists in the given subtree

SYNOPSIS
  merge_key_exists();
  root        subtree root
  elem        element to fine
  
RETURN VALUES
  int         position of the element if found
  -1          no element found
*/
int merge_key_exists(Jtree_elem* root, Jtree_elem* elem);


/* 
Converts tree to string

SYNOPSIS
  fill_result();
  jte           tree root
  result        string to fill up with the result
  total_offset  offset from current position
  
RETURN VALUES
  int         new offset
*/
int fill_result(Jtree_elem* jte, char* result, int total_offset= 0);

#endif /* MY_JSON */

/* :tabSize=2:indentSize=2:noTabs=true:folding=custom:collapseFolds=1: */
